/*
  File:SoundflowerDevice.cpp

  Version:1.0.1
    ma++ ingalls  |  cycling '74  |  Copyright (C) 2004  |  soundflower.com
    
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
    Soundflower is derived from Apple's 'PhantomAudioDriver'
    sample code.  It uses the same timer mechanism to simulate a hardware
    interrupt, with some additional code to compensate for the software
    timer's inconsistencies.  
    
    Soundflower basically copies the mixbuffer and presents it to clients
    as an input buffer, allowing applications to send audio one another.
*/

#include "SoundflowerDevice.h"
#include "SoundflowerEngine.h"
#include <IOKit/audio/IOAudioControl.h>
#include <IOKit/audio/IOAudioLevelControl.h>
#include <IOKit/audio/IOAudioToggleControl.h>
#include <IOKit/audio/IOAudioDefines.h>
#include <IOKit/IOLib.h>

#define super IOAudioDevice

OSDefineMetaClassAndStructors(SoundflowerDevice, IOAudioDevice)

// There should probably only be one of these? This needs to be 
// set to the last valid position of the log lookup table. 
const SInt32 SoundflowerDevice::kVolumeMax = 99;
const SInt32 SoundflowerDevice::kGainMax = 99;

int com_nomachine_nxaudio_connected = 0;

IOAudioDevice *com_nomachine_nxaudio_device = NULL;

kern_ctl_ref com_nomachine_nxaudio_kctlref = NULL;


bool SoundflowerDevice::initHardware(IOService *provider)
{
    bool result = false;
    
	//IOLog("SoundflowerDevice[%p]::initHardware(%p)\n", this, provider);
    
    if (!super::initHardware(provider))
        goto Done;
    
    setDeviceName("nxaudio");
    setDeviceShortName("nxaudio");
    setManufacturerName("Soundflower - ma++ ingalls for Cycling '74");

    setDeviceCanBeDefault(kIOAudioDeviceCanBeDefaultNothing);

    com_nomachine_nxaudio_device = this;

    initKernControl();
    
    if (!createAudioEngines())
        goto Done;
    
    result = true;
    
Done:

    return result;
}


bool SoundflowerDevice::createAudioEngines()
{
    OSArray*				audioEngineArray = OSDynamicCast(OSArray, getProperty(AUDIO_ENGINES_KEY));
    OSCollectionIterator*	audioEngineIterator;
    OSDictionary*			audioEngineDict;
	
    if (!audioEngineArray) {
        IOLog("SoundflowerDevice[%p]::createAudioEngine() - Error: no AudioEngine array in personality.\n", this);
        return false;
    }
    
	audioEngineIterator = OSCollectionIterator::withCollection(audioEngineArray);
    if (!audioEngineIterator) {
		IOLog("SoundflowerDevice: no audio engines available.\n");
		return true;
	}
    
    while (audioEngineDict = (OSDictionary*)audioEngineIterator->getNextObject()) {
		SoundflowerEngine*	audioEngine = NULL;
		
        if (OSDynamicCast(OSDictionary, audioEngineDict) == NULL)
            continue;
        
		audioEngine = new SoundflowerEngine;
        if (!audioEngine)
			continue;
        
        if (!audioEngine->init(audioEngineDict))
			continue;

		initControls(audioEngine);
        activateAudioEngine(audioEngine);	// increments refcount and manages the object
        audioEngine->release();				// decrement refcount so object is released when the manager eventually releases it
    }
	
    audioEngineIterator->release();
    return true;
}


#define addControl(control, handler) \
    if (!control) {\
		IOLog("Soundflower failed to add control.\n");	\
		return false; \
	} \
    control->setValueChangeHandler(handler, this); \
    audioEngine->addDefaultAudioControl(control); \
    control->release();

bool SoundflowerDevice::initControls(SoundflowerEngine* audioEngine)
{
    IOAudioControl*	control = NULL;
    
    for (UInt32 channel=0; channel <= 16; channel++) {
        mVolume[channel] = mGain[channel] = kVolumeMax;
        mMuteOut[channel] = mMuteIn[channel] = false;
    }
    
    const char *channelNameMap[17] = {	kIOAudioControlChannelNameAll,
										kIOAudioControlChannelNameLeft,
										kIOAudioControlChannelNameRight,
										kIOAudioControlChannelNameCenter,
										kIOAudioControlChannelNameLeftRear,
										kIOAudioControlChannelNameRightRear,
										kIOAudioControlChannelNameSub };
	
    for (UInt32 channel=7; channel <= 16; channel++)
        channelNameMap[channel] = "Unknown Channel";
    
    for (unsigned channel=0; channel <= 16; channel++) {
		
        // Create an output volume control for each channel with an int range from 0 to 65535
        // and a db range from -72 to 0
        // Once each control is added to the audio engine, they should be released
        // so that when the audio engine is done with them, they get freed properly
		 
		 // Volume level notes.
		 //
		 // Previously, the minimum volume was actually 10*log10(1/65535) = -48.2 dB. In the new
		 // scheme, we use a size 100 lookup table to compute the correct log scaling. And set
		 // the minimum to -40 dB. Perhaps -50 dB would have been better, but this seems ok.
		 
        control = IOAudioLevelControl::createVolumeControl(SoundflowerDevice::kVolumeMax,		// Initial value
                                                           0,									// min value
                                                           SoundflowerDevice::kVolumeMax,		// max value
                                                           (-40 << 16) + (32768),				// -72 in IOFixed (16.16)
                                                           0,									// max 0.0 in IOFixed
                                                           channel,								// kIOAudioControlChannelIDDefaultLeft,
                                                           channelNameMap[channel],				// kIOAudioControlChannelNameLeft,
                                                           channel,								// control ID - driver-defined
                                                           kIOAudioControlUsageOutput);
        addControl(control, (IOAudioControl::IntValueChangeHandler)volumeChangeHandler);
        
        // Gain control for each channel
        control = IOAudioLevelControl::createVolumeControl(SoundflowerDevice::kGainMax,			// Initial value
                                                           0,									// min value
                                                           SoundflowerDevice::kGainMax,			// max value
                                                           0,									// min 0.0 in IOFixed
                                                           (40 << 16) + (32768),				// 72 in IOFixed (16.16)
                                                           channel,								// kIOAudioControlChannelIDDefaultLeft,
                                                           channelNameMap[channel],				// kIOAudioControlChannelNameLeft,
                                                           channel,								// control ID - driver-defined
                                                           kIOAudioControlUsageInput);
        addControl(control, (IOAudioControl::IntValueChangeHandler)gainChangeHandler);
    }
	
    // Create an output mute control
    control = IOAudioToggleControl::createMuteControl(false,									// initial state - unmuted
                                                      kIOAudioControlChannelIDAll,				// Affects all channels
                                                      kIOAudioControlChannelNameAll,
                                                      0,										// control ID - driver-defined
                                                      kIOAudioControlUsageOutput);
    addControl(control, (IOAudioControl::IntValueChangeHandler)outputMuteChangeHandler);
    
    // Create an input mute control
    control = IOAudioToggleControl::createMuteControl(false,									// initial state - unmuted
                                                      kIOAudioControlChannelIDAll,				// Affects all channels
                                                      kIOAudioControlChannelNameAll,
                                                      0,										// control ID - driver-defined
                                                      kIOAudioControlUsageInput);
    addControl(control, (IOAudioControl::IntValueChangeHandler)inputMuteChangeHandler);
    
    return true;
}


IOReturn SoundflowerDevice::volumeChangeHandler(IOService *target, IOAudioControl *volumeControl, SInt32 oldValue, SInt32 newValue)
{
    IOReturn			result = kIOReturnBadArgument;
    SoundflowerDevice*	audioDevice = (SoundflowerDevice *)target;
	
    if (audioDevice)
        result = audioDevice->volumeChanged(volumeControl, oldValue, newValue);
    return result;
}


IOReturn SoundflowerDevice::volumeChanged(IOAudioControl *volumeControl, SInt32 oldValue, SInt32 newValue)
{
    if (volumeControl)
         mVolume[volumeControl->getChannelID()] = newValue;
    return kIOReturnSuccess;
}


IOReturn SoundflowerDevice::outputMuteChangeHandler(IOService *target, IOAudioControl *muteControl, SInt32 oldValue, SInt32 newValue)
{
    IOReturn			result = kIOReturnBadArgument;
    SoundflowerDevice*	audioDevice = (SoundflowerDevice*)target;
	
    if (audioDevice)
        result = audioDevice->outputMuteChanged(muteControl, oldValue, newValue);
    return result;
}


IOReturn SoundflowerDevice::outputMuteChanged(IOAudioControl *muteControl, SInt32 oldValue, SInt32 newValue)
{
    if (muteControl)
         mMuteOut[muteControl->getChannelID()] = newValue;
    return kIOReturnSuccess;
}


IOReturn SoundflowerDevice::gainChangeHandler(IOService *target, IOAudioControl *gainControl, SInt32 oldValue, SInt32 newValue)
{
    IOReturn			result = kIOReturnBadArgument;
    SoundflowerDevice*	audioDevice = (SoundflowerDevice *)target;
	
    if (audioDevice)
        result = audioDevice->gainChanged(gainControl, oldValue, newValue);
    return result;
}


IOReturn SoundflowerDevice::gainChanged(IOAudioControl *gainControl, SInt32 oldValue, SInt32 newValue)
{
    if (gainControl)
		mGain[gainControl->getChannelID()] = newValue;
    return kIOReturnSuccess;
}


IOReturn SoundflowerDevice::inputMuteChangeHandler(IOService *target, IOAudioControl *muteControl, SInt32 oldValue, SInt32 newValue)
{
    IOReturn			result = kIOReturnBadArgument;
    SoundflowerDevice*	audioDevice = (SoundflowerDevice*)target;

    if (audioDevice)
        result = audioDevice->inputMuteChanged(muteControl, oldValue, newValue);
    return result;
}

IOReturn SoundflowerDevice::inputMuteChanged(IOAudioControl *muteControl, SInt32 oldValue, SInt32 newValue)
{
    if (muteControl)
         mMuteIn[muteControl->getChannelID()] = newValue;
    return kIOReturnSuccess;
}

void SoundflowerDevice::initKernControl()
{
    errno_t error;

    struct kern_ctl_reg ep_ctl;
 
    bzero(&ep_ctl, sizeof(ep_ctl));

    ep_ctl.ctl_id   = 0;
    ep_ctl.ctl_unit = 0;

    strncpy(ep_ctl.ctl_name, AUDIO_CONTROL_NAME, MAX_KCTL_NAME);

    ep_ctl.ctl_flags = 0;

    ep_ctl.ctl_send       = handleWriteControl;
    ep_ctl.ctl_getopt     = handleGetControl;
    ep_ctl.ctl_setopt     = handleSetControl;
    ep_ctl.ctl_connect    = handleConnectControl;
    ep_ctl.ctl_disconnect = handleDisconnectControl;

    error = ctl_register(&ep_ctl, &com_nomachine_nxaudio_kctlref);
}
 
errno_t SoundflowerDevice::handleSetControl(kern_ctl_ref ctlref, unsigned int unit, void *userdata, int opt, void *data, size_t len )
{
    return (0);
}

errno_t SoundflowerDevice::handleGetControl(kern_ctl_ref ctlref, unsigned int unit, void *userdata, int opt, void *data, size_t *len)
{
    return (0);
}

errno_t SoundflowerDevice::handleConnectControl(kern_ctl_ref ctlref, struct sockaddr_ctl *sac, void **unitinfo)
{
    com_nomachine_nxaudio_connected++;

    if (com_nomachine_nxaudio_connected == 1)
    {
        if (com_nomachine_nxaudio_device != NULL)
        {
            com_nomachine_nxaudio_device -> setDeviceCanBeDefault(kIOAudioDeviceCanBeDefaultInput |
                                                kIOAudioDeviceCanBeDefaultOutput |
                                                    kIOAudioDeviceCanBeSystemOutput);
        }
    }

    return (0);
}

errno_t SoundflowerDevice::handleDisconnectControl(kern_ctl_ref ctlref, unsigned int unit, void *unitinfo)
{
    com_nomachine_nxaudio_connected--;

    if (com_nomachine_nxaudio_connected == 0)
    {
        if (com_nomachine_nxaudio_device != NULL)
        {
            com_nomachine_nxaudio_device -> setDeviceCanBeDefault(kIOAudioDeviceCanBeDefaultNothing);
        }
    }

    return (0);
}

errno_t SoundflowerDevice::handleWriteControl(kern_ctl_ref ctlref, unsigned int unit, void *userdata, mbuf_t m, int flags)
{
    return (0);
}

bool SoundflowerDevice::terminate(IOOptionBits options)
{
  	// IOLog("nxaudio::terminate()\n");

    if (com_nomachine_nxaudio_kctlref != NULL)
    {
        if (ctl_deregister(com_nomachine_nxaudio_kctlref) != 0)
        {
  	        // IOLog("nxaudio::terminate(): Can not deregister control.\n");

            return false;
        }
    }

  	// IOLog("nxaudio::terminate(): Terminating.\n");

    return super::terminate(options);
}

