/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "AudioAdapter.h"

#pragma code_seg("PAGE")

//
// Wrapped constructor for AudioAdapter needed for driver 
// initialization code (-> see 'Init.cpp').
//
// unknown  - pointer to new create adapter (OUT)
// outer    - ? (IN)
// poolType - pool type to use (IN)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS CreateAudioAdapter(PUNKNOWN *unknown, REFCLSID,
                                 PUNKNOWN outer, POOL_TYPE poolType)
{
  PAGED_CODE();

  DBG_MSG("CreateAudioAdapter()...\n");

  ASSERT(unknown);
  
  STD_CREATE_BODY_(AudioAdapter, unknown, outer, poolType, AudioAdapter *);
}


//
// Destructor release used interfaces (see -> COM model).
//

AudioAdapter::~AudioAdapter()
{
  PAGED_CODE();

  DBG_MSG("~AudioAdapter()...\n");

  if (portWaveCyclic_)
  {
    portWaveCyclic_ -> Release();
  }

  if (serviceGroup_)
  {
    serviceGroup_ -> Release();
  }
}

#pragma code_seg()

//
// Returns device object related with this AudioAdapter.
//

DEVICE_OBJECT *AudioAdapter::getDeviceObject()
{
  return device_;
}

#pragma code_seg("PAGE")

//
// Initialize AudioAdapter.
//
// device - device object related with adapter (IN)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS AudioAdapter::init(DEVICE_OBJECT *device)
{
  PAGED_CODE();

  ASSERT(device);

  DBG_MSG("AudioAdapter::init()...\n");

  device_ = device;
 
  powerState_ = PowerDeviceD0;
  
  mixerReset();

  return STATUS_SUCCESS;
}


//
// Reset property values.
//

void AudioAdapter::mixerReset()
{
  PAGED_CODE();
    
  RtlFillMemory(volume_, sizeof(LONG) * MAX_TOPOLOGY_NODES, 0xFF);

  RtlFillMemory(mute_, sizeof(BOOL) * MAX_TOPOLOGY_NODES, FALSE);
 
  mux_ = 2;
}


//
// Query one of interface implemented by class (-> see COM model)
// and increese its reference counter if success.
//
// id     - interface id (IN)
// object - pointer to quered interface (OUT)
//
// RETURNS: NTSTATUS code.
//

STDMETHODIMP AudioAdapter::NonDelegatingQueryInterface(REFIID id, PVOID *object)
{
  PAGED_CODE();

  ASSERT(object);

  if (id == IID_IUnknown)
  {
    *object = (IUnknown *) this;
  }
  else if (id == IID_AudioAdapter)
  {
    *object = (AudioAdapter *) this;
  }
  else if (id == IID_IAdapterPowerManagement)
  {
    *object = (PADAPTERPOWERMANAGEMENT) this;
  }
  else
  {
    *object = NULL;
  }

  if (*object)
  {
    PUNKNOWN(*object) -> AddRef();

    return STATUS_SUCCESS;
  }

  return CUnknown::NonDelegatingQueryInterface(id, object);
}

//
// Set service group object needed for sample rate synchronization.
//

void AudioAdapter::setWaveServiceGroup(PSERVICEGROUP serviceGroup)
{
  PAGED_CODE();
  
  DBG_MSG("AudioAdapter::setWaveServiceGrou()...\n");
  
  if (serviceGroup_)
  {
    serviceGroup_ -> Release();
  }

  serviceGroup_ = serviceGroup;

  if (serviceGroup_)
  {
    serviceGroup_ -> AddRef();
  }
}

//
// Returns mother WavePortCyclic object. (Our MiniPort driver is
// wrapped by WavePortCyclic).
//

PUNKNOWN *AudioAdapter::getWavePortCyclic()
{
  PAGED_CODE();

  return (PUNKNOWN *) &portWaveCyclic_;
}


//
// Setters and getters for propery values.
//

BOOL AudioAdapter::getMute(ULONG node)
{
  if (node < MAX_TOPOLOGY_NODES)
  {
    return mute_[node];
  }

  return 0;
}


void AudioAdapter::setMute(ULONG node, BOOL state)
{
  if (node < MAX_TOPOLOGY_NODES)
  {
    mute_[node] = state;
  }
}


ULONG AudioAdapter::getMux()
{
  return mux_;
}


void AudioAdapter::setMux(ULONG node)
{
  mux_ = node;
}


LONG AudioAdapter::getVolume(ULONG node)
{
  if (node < MAX_TOPOLOGY_NODES)
  {
    return volume_[node];
  }

  return 0;
}


void AudioAdapter::setVolume(ULONG node, LONG volume)
{
  if (node < MAX_TOPOLOGY_NODES)
  {
    volume_[node] = volume;
  }
}


//
// Below functions implement IAdapterPowerManagement interface.
//

STDMETHODIMP_(void) AudioAdapter::PowerChangeState(POWER_STATE NewState)
{
  DBG_MSG("AudioAdapter::powerChangeState()...\n");

  //
  // is this actually a state change??
  //

  if (NewState.DeviceState != powerState_)
  {
    //
    // switch on new state
    //
    
    switch (NewState.DeviceState)
    {
      case PowerDeviceD0:
      case PowerDeviceD1:
      case PowerDeviceD2:
      case PowerDeviceD3:
      {
        powerState_ = NewState.DeviceState;

        break;
      }  
  
      default:
      {
        break;
      }  
    }
  }
}

STDMETHODIMP_(NTSTATUS) AudioAdapter::QueryDeviceCapabilities(PDEVICE_CAPABILITIES PowerDeviceCaps)
{
  UNREFERENCED_PARAMETER(PowerDeviceCaps);

  DBG_MSG("AudioAdapter::QueryDeviceCapabilities()...\n");

  return (STATUS_SUCCESS);
}


STDMETHODIMP_(NTSTATUS) AudioAdapter::QueryPowerChangeState(POWER_STATE NewStateQuery)
{
  UNREFERENCED_PARAMETER(NewStateQuery);

  DBG_MSG("AudioAdapter::QueryPowerChangeState()...\n");

  return STATUS_SUCCESS;
}

#pragma code_seg()
