/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#define PUT_GUIDS_HERE

#include "Init.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// Driver entry point. We pass pointer to 'AddDevice' function here.
//
// driver   - driver object given from system (IN)
// registry - registry path for driver info given by system (IN)
//
// RETURNS: NTSTATUS code.
//

#pragma code_seg("INIT")

NTSTATUS DriverEntry(PDRIVER_OBJECT driver, PUNICODE_STRING registry)
{
  DBG_MSG("DriverEntry()[6.8.1]...\n");
  
  return PcInitializeAdapterDriver(driver, registry, 
                                       (PDRIVER_ADD_DEVICE) AddDevice);
}
#pragma code_seg()

#pragma code_seg("PAGE")


//
// Add device related to given driver. We pass pointer to
// StartDevice function here.
//
// driver - driver object given by system in DriverEntry (IN)
// device - device object given by system (IN)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS AddDevice(PDRIVER_OBJECT driver, PDEVICE_OBJECT device)
{
  PAGED_CODE();

  DBG_MSG("AddDevice()...\n");
  
  return PcAddAdapterDevice(driver, device, 
                                PCPFNSTARTDEVICE(StartDevice), 
                                    MAX_MINIPORTS, 0);
}


//  
// This function is called by system when device is started.
// We create one common AudioAdapter object here and add
// two subdevices to it: Topology and Wave.
//
// device   - device driver object (IN)
// irp      - pointer to the irp object (IN) (?)
// resource - resource list assigned by PnP manager (IN)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS StartDevice(PDEVICE_OBJECT device, PIRP irp, PRESOURCELIST resources)
{
  DBG_MSG("-> StartDevice()...\n");

  UNREFERENCED_PARAMETER(resources);

  PAGED_CODE();
  
  NTSTATUS ntStatus = STATUS_UNSUCCESSFUL;

  PUNKNOWN unknownTopology = NULL;
  PUNKNOWN unknownWave     = NULL;
  PUNKNOWN unknownCommon   = NULL;
  
  AudioAdapter *audioAdapter = NULL;


  //
  // create a new adapter common object.
  //
  
  NTFAIL(CreateAudioAdapter(&unknownCommon,
                                IID_AudioAdapter, NULL, NonPagedPool));

  NTFAIL(unknownCommon -> QueryInterface(IID_AudioAdapter,
                                              (PVOID *) &audioAdapter));

  NTFAIL(audioAdapter -> init(device));

  //
  // register with PortCls for power-management services.
  //    
        
  NTFAIL(PcRegisterAdapterPowerManagement(PUNKNOWN(audioAdapter), device));

  //
  //  install Topology subdevice.
  //
  
  NTFAIL(InstallSubdevice(device, irp, L"Topology", CLSID_PortTopology,
                              CLSID_PortTopology, CreateMiniPortTopology,
                                  audioAdapter, NULL, IID_IPortTopology,
                                      NULL, &unknownTopology));

  //
  // install Wave subdevice.
  //

  NTFAIL(InstallSubdevice(device, irp, L"Wave", CLSID_PortWaveCyclic,
                              CLSID_PortWaveCyclic, CreateMiniportWaveCyclic,
                                  audioAdapter, NULL, IID_IPortWaveCyclic,
                                      audioAdapter -> getWavePortCyclic(),
                                          &unknownWave));

  //
  // Register connections beetwen Wave and Topology subdevices.
  //

  if (unknownTopology && unknownWave)
  {
    ULONG waveIn  = Topology::physicalConnections_.waveIn_;
    ULONG waveOut = Topology::physicalConnections_.waveOut_;

    ULONG topologyIn  = Topology::physicalConnections_.topologyIn_;
    ULONG topologyOut = Topology::physicalConnections_.topologyOut_;

    if (topologyOut != -1 && waveIn != -1)
    {
      NTFAIL(PcRegisterPhysicalConnection(device, unknownTopology, 
                                              topologyOut, unknownWave, waveIn));
    }
    
    if (waveOut != -1 && topologyIn != -1)
    {
      NTFAIL(PcRegisterPhysicalConnection(device, unknownWave, waveOut, 
                                              unknownTopology, topologyIn));
    }    
  }

  ntStatus = STATUS_SUCCESS;

fail:

  //
  // Clean up.
  //
  
  if (audioAdapter)
  {
    audioAdapter -> Release();
  }

  if (unknownCommon)
  {
    unknownCommon -> Release();
  }
  
  if (unknownTopology)
  {
    unknownTopology -> Release();
  }

  if (unknownWave)
  {
    unknownWave -> Release();
  }  
  
  DBG_MSG("<- StartDevice()...\n");
  
  return ntStatus;
}


//
//
//
//

NTSTATUS InstallSubdevice(PDEVICE_OBJECT device, PIRP irp,
                              PWSTR name, REFGUID portClassId,
                                  REFGUID miniportClassId,
                                      PFNCREATEINSTANCE miniportCreate,
                                          PUNKNOWN adapter,
                                              PRESOURCELIST resource,
                                                  REFGUID portInterfaceId,
                                                      PUNKNOWN *outPortInterface,
                                                          PUNKNOWN *outPortUnknown)
{
  PAGED_CODE();

  NTSTATUS ntStatus = STATUS_INTERNAL_ERROR;

  PPORT port = NULL;
  
  PUNKNOWN miniport = NULL;


  DBG_MSG("-> InstallSubDevice(%ls)...\n", name);

  //
  // Create the port driver object
  //
  
  NTFAIL(PcNewPort(&port, portClassId));

  //
  // Create the miniport object
  //
  
  if (miniportCreate)
  {
    NTFAIL(miniportCreate(&miniport, miniportClassId, NULL, NonPagedPool));
  }
  else
  {
    NTFAIL(PcNewMiniport((PMINIPORT *) &miniport, miniportClassId));
  }

  //
  // Init the port driver and miniport in one go.
  //
  
  NTFAIL(port -> Init(device, irp, miniport, adapter, resource));

  //
  // Register the subdevice (port/miniport combination).
  //

  NTFAIL(PcRegisterSubdevice(device, name, port));

  //
  // We don't need the miniport any more.  Either the port has it,
  // or we've failed, and it should be deleted.
  //
  
  //
  // Deposit the port interfaces if it's needed.
  //
  
  if (outPortUnknown)
  {
    NTFAIL(port -> QueryInterface(IID_IUnknown, (PVOID *) outPortUnknown));
  
    if (outPortInterface)
    {
      NTFAIL(port -> QueryInterface(portInterfaceId, (PVOID *) outPortInterface));
    }
  }

  ntStatus = STATUS_SUCCESS;
  
fail:

  //
  // Clean up.
  //
  
  if (port)
  {
    port -> Release();
  }
  
  if (miniport)
  {
    miniport -> Release();
  }
  
  DBG_MSG("<- InstallSubDevice(%ls)...\n", name);
  
  return ntStatus;
}

#pragma code_seg()

#ifdef __cplusplus
}
#endif
