/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "MiniPort.h"

MiniPort::MiniPort(PUNKNOWN unknown) : CUnknown(unknown)
{
  audioAdapter_          = NULL;
  portWaveCyclic_        = NULL;

  serviceGroup_          = NULL;
  notifyInterval_        = 0;

  buffer_                = NULL;
  bufferSize_            = 0;
  writePos_              = 0;
  readPos_               = 0;

  writePosLap_           = 0;
  readPosLap_            = 0;
  availableBufferToRead_ = 0;

  channels_              = 0;
  resolution_            = 0;
  sampleRate_            = 0;
}

//
// Create new MiniPort object.
//

NTSTATUS CreateMiniportWaveCyclic(PUNKNOWN *Unknown, REFCLSID, 
                                      PUNKNOWN UnknownOuter OPTIONAL,
                                          POOL_TYPE PoolType)
{
  DBG_MSG("CreateMiniportWaveCyclic()...\n");

  ASSERT(Unknown);

  STD_CREATE_BODY(MiniPort, Unknown, UnknownOuter, PoolType);
}

//
// Initialize new MiniPort driver. We only set up our internal 
// sample format here.
//

STDMETHODIMP_(NTSTATUS) MiniPort::Init(PUNKNOWN unknownAdapter,
                                           PRESOURCELIST resourceList,
                                               PPORTWAVECYCLIC port)
{
  UNREFERENCED_PARAMETER(resourceList);

  NTSTATUS ntStatus = STATUS_SUCCESS;

  DBG_MSG("MiniPort::Init()...\n");

  //
  // Initialize members.
  //

  sampleRate_     = 0;
  notifyInterval_ = 0;

  isRecording_ = false;
  isPlaying_   = false;

  //
  // Add reference to used interfaces (PortWaveCyclic, audioAdapter).
  //
  
  portWaveCyclic_ = port;

  portWaveCyclic_ -> AddRef();

  NTFAIL(unknownAdapter -> QueryInterface(IID_AudioAdapter,
                                              (PVOID *) &audioAdapter_));

  //
  // Init mutex and serviceGroup used for sample rate synchronizing.
  //
  
  KeInitializeMutex(&sampleRateSync_, 0);
    
  NTFAIL(PcNewServiceGroup(&serviceGroup_, NULL));

  audioAdapter_ -> setWaveServiceGroup(serviceGroup_);
  
  //
  // Set up our internal sample format.
  //

  channels_   = INTERNAL_CHANNELS;
  resolution_ = INTERNAL_RESOLUTION;
  sampleRate_ = INTERNAL_SAMPLE_RATE;

  //
  // Allocate common Buffer for record and play.
  //
  
  bufferSize_ = BUFFER_SIZE;
  
  if (buffer_ == NULL)
  {
    DBG_MSG("MiniPort::Init() : Alocating common buffer...\n");
    
    buffer_ = (BYTE *) ExAllocatePoolWithTag(NonPagedPool, bufferSize_, NX_POOLTAG);

    if (!buffer_)
    {
      ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }
    else
    {
      writePos_ = bufferSize_ / 2;

      readPos_  = 0;
    }
  }

  if (NT_SUCCESS(ntStatus))
  {
    availableBufferToRead_ = 0;

    writePosLap_ = false;

    readPosLap_  = false;
  }

fail:
 
  //
  // Clean up if error.
  //
  
  if (NT_ERROR(ntStatus))
  {
    DBG_MSG("ERROR. Cannot init MiniPort object (ntStat = %x).\n", ntStatus);
    
    if (audioAdapter_)
    {
      //
      // Clean up the service group.
      //
      
      if (serviceGroup_)
      {
        audioAdapter_ -> setWaveServiceGroup(NULL);

        serviceGroup_ -> Release();
        
        serviceGroup_ = NULL;
      }

      audioAdapter_ -> Release();

      audioAdapter_ = NULL;
    }

    //
    // release the port
    //
    
    portWaveCyclic_ -> Release();

    portWaveCyclic_ = NULL;
  }
  
  return ntStatus;
}

//
// Creates new stream for recording or playing.
//
// outStream       - pointer to new created stream. (OUT)
// outerUnknown    - ?
// poolType        - ?
// pin             - Wave Topology Pin asociated with stream. (IN)
// capture         - true for record streams. (IN)
// dataFormat      - format of new stream. (IN)
//
// outDmaChannel   - pointer to DmaChannel associated with stream,
//                   we can return stream object becouse Stream 
//                   derivates from DmaChannel class. (OUT)
//
// outServiceGroup - ?
//
// RETURNS: STATUS_SUCCESS if O.K.
//

STDMETHODIMP_(NTSTATUS)
  MiniPort::NewStream(PMINIPORTWAVECYCLICSTREAM *outStream, PUNKNOWN outerUnknown,
                          POOL_TYPE poolType, ULONG pin, BOOLEAN capture,
                              PKSDATAFORMAT dataFormat, PDMACHANNEL *outDmaChannel,
                                  PSERVICEGROUP *outServiceGroup)
{
  UNREFERENCED_PARAMETER(poolType);

  DBG_MSG("-> MiniPort::NewStream()...\n");

  ASSERT(outStream);
  ASSERT(dataFormat);
  ASSERT(outDmaChannel);
  ASSERT(outServiceGroup);

  NTSTATUS ntStatus = STATUS_SUCCESS;

  Stream *stream = NULL;

  //
  // We can record/play only one stream. It's NOT number of samples, 
  // which we can play/record at once, becouse KMIXER mixes streams
  // before deliver they to us.
  //

  if (capture)
  {
    if (isRecording_)
    {
      DBG_MSG("ERROR: Only one recording stream supported.\n");

      return STATUS_INSUFFICIENT_RESOURCES;
    }
  }
  else
  {
    if (isPlaying_)
    {
      DBG_MSG("ERROR: Only one playing stream supported.\n");

      return STATUS_INSUFFICIENT_RESOURCES;
    }
  }
  
  // 
  // If recording and playing are active together, the goal is to
  // set read position little before write position. It makes extra
  // lattency on record stream, but then reading data from buffer
  // is more stable due to extra distance beetwen ReadPos and WritePos.
  //

  //
  // new recording stream
  //
  
  if (capture)
  {
    if (isPlaying_)
    {
      readPos_ = writePos_ - bufferSize_ / 2;
      
      if (readPos_ < 0)
      {
        readPos_ = readPos_ + bufferSize_;
      } 

      if (readPos_ > (int) bufferSize_)
      {
        readPos_ = 0;
      }
    }
    else
    {
      readPos_ = 0; 
    }
  }

  else
 
  //
  // new playback stream
  //
  
  {
    if (isRecording_)
    {
      writePos_ = (readPos_ + bufferSize_ / 2) % bufferSize_;
    }
    else
    {
      writePos_ = bufferSize_ /2;
    }
  }
  
  //
  // Is proposed format correct? We use function from base driver.
  //

  if (NT_SUCCESS(ntStatus))
  {
    ntStatus = checkFormat(dataFormat);
  }

  //
  // Create new stream object
  //

  if (NT_SUCCESS(ntStatus))
  {
    stream = new (NonPagedPool, NX_POOLTAG) Stream(outerUnknown);

    if (stream)
    {
      stream -> AddRef();

      ntStatus = stream -> init(this, pin, capture, dataFormat);
    }
    else
    {
      ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }
  }

  if (NT_SUCCESS(ntStatus))
  {
    //
    // Update recording/playing flag.
    //

    if (capture)
    {
      isRecording_ = true;
    }
    else
    {
      isPlaying_ = true;
    }

    //
    // Fill output parameter with created stream.
    //

    *outStream = PMINIPORTWAVECYCLICSTREAM(stream);

    (*outStream) -> AddRef();

    *outDmaChannel = PDMACHANNEL(stream);

    (*outDmaChannel) -> AddRef();

    *outServiceGroup = serviceGroup_;

    (*outServiceGroup) -> AddRef();
  }

  //
  // Error message if fail.
  //
  
  if (ntStatus)
  {
    DBG_MSG("ERROR. MiniPort::NewStream() fails (ntStat = %x).\n", ntStatus);
  }
  
  //
  // Clean up.
  //
  
  if (stream)
  {
    stream -> Release();
  }

  DBG_MSG("<- MiniPort::NewStream()...\n");
  
  return ntStatus;
}


//
// Get request for ComponentId driver property from client application.
// We fill request struct with data unique for our driver here.
//
// request - struct, which descripts reqeust (IN/OUT)
// 
// RETURNS: STATUS_SUCCESS if OK.
//

NTSTATUS MiniPort::propertyHandlerComponentId(PPCPROPERTY_REQUEST request)
{
  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;
  
  DBG_MSG("MiniPort::PropertyHandlerComponentId()...\n");

  //
  // Check arguments.
  //
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);
  
  //
  // Basic support request.
  //

  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    FAIL(request -> ValueSize < sizeof(ULONG));

    *(PULONG(request -> Value)) = KSPROPERTY_TYPE_BASICSUPPORT | KSPROPERTY_TYPE_GET;

    request -> ValueSize = sizeof(ULONG);
  }

  //
  // Get component ID property , we return infos unique for our driver here.
  //
    
  else if (request -> Verb & KSPROPERTY_TYPE_GET)
  {
    FAIL(request -> ValueSize < sizeof(KSCOMPONENTID));

    KSCOMPONENTID *componentId = (KSCOMPONENTID *) request -> Value;

    componentId -> Manufacturer = NXAUDIO_MANUFACTURER;
    componentId -> Product      = NXAUDIO_PID;

    //
    // FIXME. We don't pass GUID for Device's name longer, becouse
    // we want to have "Audio Adapter" device one time, and
    // "Microphone Adapter" another time. If we pass NULL guid
    // here system tries to use friendly names of interfaces
    // from INF file.
    //
    // componentId -> Name         = NXAUDIO_NAME;
    //
    
    componentId -> Name         = GUID_NULL;
    componentId -> Component    = GUID_NULL;
    componentId -> Version      = NXAUDIO_VERSION;
    componentId -> Revision     = NXAUDIO_REVISION;

    request -> ValueSize = sizeof(KSCOMPONENTID);
  }
  else
  {
    NTFAIL(STATUS_INVALID_DEVICE_REQUEST);
  }
  
  ntStatus = STATUS_SUCCESS;

fail:

  if (ntStatus)
  {
    DBG_MSG("ERROR. MiniPort::propertyHandlerComponentId() fails (ntStat = %x).\n", ntStatus);
  }

  return ntStatus;
}

//
// ProposedFormat request from client aplication.
//
// request - struct, which descript proposed format (IN)
//
// RETURNS: STATUS_SUCCESS if we support format. 
//

NTSTATUS MiniPort::propertyHandlerProposedFormat(PPCPROPERTY_REQUEST request)
{
  DBG_MSG("MiniPort::propertyHandlerProposedFormat()...\n");
  
  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;

  //
  // Basic supoprt request.
  //

  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    FAIL(request -> ValueSize < sizeof(ULONG));

    *(PULONG(request -> Value)) = KSPROPERTY_TYPE_BASICSUPPORT | KSPROPERTY_TYPE_SET;

    request -> ValueSize = sizeof(ULONG);
  }

  //
  // Set format property.
  //

  else if (request -> Verb & KSPROPERTY_TYPE_SET)
  {
    //
    // We support only PCM AUDIO format. Format details (sampe rate
    // etc.) doesn't matter for us, becouse we don't do anything
    // format specified. KMIXER makes format conversion before our
    // driver if needed.
    //

    KSDATAFORMAT_WAVEFORMATEX *format = (KSDATAFORMAT_WAVEFORMATEX *) request -> Value;
        
    FAIL(format -> DataFormat.MajorFormat != KSDATAFORMAT_TYPE_AUDIO);
    FAIL(format -> DataFormat.SubFormat != KSDATAFORMAT_SUBTYPE_PCM);
  }
  else
  {
    NTFAIL(STATUS_INVALID_DEVICE_REQUEST);
  }
  
  ntStatus = STATUS_SUCCESS;  
   
fail:

  if (ntStatus)
  {
    DBG_MSG("MiniPort::propertyHandlerProposedFormat() fails (ntStat = %x)...\n", ntStatus);
  }
  
  return ntStatus;
}

//
// Handler for all MiniPort property requests.
//
// request - structure describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS MiniPort::propertyHandler(PPCPROPERTY_REQUEST request)
{
  MiniPort *miniPort = (MiniPort *) request -> MajorTarget;
  
  DBG_MSG("MiniPort::propertyHandler(id = %u)...\n", request -> PropertyItem -> Id);
  
  //
  // Custom NX property set.
  //

  if (memcmp(request -> PropertyItem -> Set, &KSPROPSETID_nxaudio, sizeof(GUID)) == 0)
  {
    return miniPort -> NxPropertyHandler(request);
  }
  
  //
  // Other standarized sets.
  //
  
  switch (request -> PropertyItem -> Id)
  {
    //
    // Component Id.
    //

    case KSPROPERTY_GENERAL_COMPONENTID:
    {
      return miniPort -> propertyHandlerComponentId(request);
    }
    
    //
    // Proposed data format.
    //
  
    case KSPROPERTY_PIN_PROPOSEDATAFORMAT:
    {
      return miniPort -> propertyHandlerProposedFormat(request);
    }

    //
    // Uknknown device request.
    //
    
    default:
    {
      DBG_MSG("MiniPort::PropertyHandler : Unknown property id.\n");

      return STATUS_INVALID_DEVICE_REQUEST;
    }  
  }
}

//
// Handler for custom NX property set requests.
//
// WARNING: We don't care about basic info for these requests.
// We assume that, only NX knows this property set (GUID) and NX
// knows how to use them.
//
// request - structure describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS MiniPort::NxPropertyHandler(PPCPROPERTY_REQUEST request)
{
  NTSTATUS ntStatus = STATUS_INVALID_DEVICE_REQUEST;
  
  DBG_MSG("MiniPort::propertyHandler(id = %u)...\n", request -> PropertyItem -> Id);
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);
  
  switch (request -> PropertyItem -> Id)
  {
    //
    // Current status of nxaudio device.
    // We encode isPlaying_ and isRecording_ flags here.
    //

    case KSPROPERTY_NXAUDIO_STATUS:
    {
      DBG_MSG("MiniPort::propertyHandler : NXAUDIO_STATUS\n");
      
      DWORD status = 0;
      
      FAIL(request -> ValueSize < sizeof(DWORD));
      FAIL(request -> Verb != KSPROPERTY_TYPE_GET);
      
      //
      // Encode status.
      //
      
      if (isPlaying_)
      {
        status |= NXAUDIO_PLAY;
      }

      if (isRecording_)
      {
        status |= NXAUDIO_RECORD;
      }
      
      //
      // Fill output.
      //
      
      *(PDWORD(request -> Value)) = status;

      break; 
    }
    
    //
    // Uknknown device request.
    //
    
    default:
    {
      DBG_MSG("MiniPort::PropertyHandler : Unknown property id.\n");

      return STATUS_INVALID_DEVICE_REQUEST;
    }  
  }

  ntStatus = STATUS_SUCCESS;
 
  fail:
  
  return ntStatus;
}


//
// Destructor decrease reference counter of used interfaces (COM model).
//

MiniPort::~MiniPort(void)
{
  DBG_MSG("MiniPort::~MiniPort()...\n");
  
  if (portWaveCyclic_)
  {
    portWaveCyclic_ -> Release();
  }

  if (serviceGroup_)
  {
    serviceGroup_ -> Release();
  }

  if (audioAdapter_)
  {
    audioAdapter_ -> Release();
  }
  
  if (buffer_)
  {
    ExFreePoolWithTag(buffer_, NX_POOLTAG);
  }  
}


//
// For compatibility with base driver only.
//

STDMETHODIMP_(NTSTATUS) 
    MiniPort::DataRangeIntersection(ULONG pin, 
                                        PKSDATARANGE clientDataRange,
                                            PKSDATARANGE myDataRange, 
                                                ULONG outputBufferLength,
                                                    PVOID resultantFormat,
                                                        PULONG resultantFormatLength)
{
  DBG_MSG("MiniPort::DataRangeIntersection()...\n");

  UNREFERENCED_PARAMETER(pin);
  UNREFERENCED_PARAMETER(clientDataRange);
  UNREFERENCED_PARAMETER(myDataRange);
  UNREFERENCED_PARAMETER(outputBufferLength);
  UNREFERENCED_PARAMETER(resultantFormat);
  UNREFERENCED_PARAMETER(resultantFormatLength);

  return STATUS_NOT_IMPLEMENTED;
}

//
// Return filter descriptor table defined in 'MiniPortTable.cpp'.
//
// filterDescriptor - pointer to table (OUT)
//
// RETURNS: NTSTATUS code.
//

STDMETHODIMP_(NTSTATUS) MiniPort::GetDescription(PPCFILTER_DESCRIPTOR *filterDescriptor)
{
  DBG_MSG("MiniPort::GetDescription()...\n");
  
  *filterDescriptor = &filterDescriptor_;

  return STATUS_SUCCESS;
}


//
// Query one of interface implemented by class (-> see COM model) 
// and increese its reference counter if success.
//
// id     - interface id (IN)
// object - pointer to quered interface (OUT)
//
// RETURNS: NTSTATUS code.
//

STDMETHODIMP_(NTSTATUS) MiniPort::NonDelegatingQueryInterface(REFIID id, PVOID *object)
{
  DBG_MSG("MiniPort::NonDelegatingQueryInterface()...\n");
  
  ASSERT(object);

  if (id == IID_IUnknown)
  {
    *object = (PMINIPORTWAVECYCLIC) this;
  }
  else if (id == IID_IMiniport)
  {
    *object = (PMINIPORT) this;
  }
  else if (id == IID_IMiniportWaveCyclic)
  {
    *object = (PMINIPORTWAVECYCLIC) this;
  }
  else
  {
    *object = NULL;
  }

  //
  // Increse reference counter for given interface.
  //

  if (*object)
  {
    PUNKNOWN(*object) -> AddRef();

    return STATUS_SUCCESS;
  }

  return CUnknown::NonDelegatingQueryInterface(id, object);
}

//
// Check is given format equally to our internal format.
//
// dataFormat - data format to check (IN)
//
// RETURNS: 0 if format OK, or STATUS_INVALID_PARAMETER in other case.
//

NTSTATUS MiniPort::checkFormat(PKSDATAFORMAT format)
{
  DBG_MSG("-> MiniPort::checkFormat()...\n");

  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;

  PWAVEFORMATEX pwfx = NULL;

  if (format && IsEqualGUIDAligned(format -> MajorFormat, KSDATAFORMAT_TYPE_AUDIO) &&
          (IsEqualGUIDAligned(format -> Specifier, KSDATAFORMAT_SPECIFIER_WAVEFORMATEX) ||
              IsEqualGUIDAligned(format -> Specifier, KSDATAFORMAT_SPECIFIER_DSOUND)))
  {
    pwfx = PWAVEFORMATEX(format + 1);

    if (IsEqualGUIDAligned(format -> Specifier, KSDATAFORMAT_SPECIFIER_DSOUND))
    {
      PKSDSOUND_BUFFERDESC pwfxds;

      pwfxds = PKSDSOUND_BUFFERDESC(format + 1);

      pwfx = &pwfxds -> WaveFormatEx;
    }
  }
  else
  {
    DBG_MSG("GUID description does not match.\n");

    NTFAIL(STATUS_INVALID_PARAMETER);
  }

  //
  // We force to use format used internal by driver.
  // So, in other case we return STATUS_INVALID_PARAMETER.
  //
  
  FAIL(pwfx == NULL);
  
  FAIL(IS_VALID_WAVEFORMATEX_GUID(&format -> SubFormat) == FALSE);
  
  FAIL(EXTRACT_WAVEFORMATEX_ID(&format -> SubFormat) != WAVE_FORMAT_PCM);

  FAIL(pwfx -> wFormatTag != WAVE_FORMAT_PCM);

  FAIL(pwfx -> cbSize != 0);
  FAIL(pwfx -> nChannels != channels_);
  FAIL(pwfx -> nSamplesPerSec != sampleRate_);
  FAIL(pwfx -> wBitsPerSample != resolution_);

  ntStatus = STATUS_SUCCESS;
  
fail:

  if (ntStatus)
  {
    DBG_MSG("MiniPort::ValidateFormat() : Invalid format.\n");
  }
  
  DBG_MSG("<- MiniPort::checkFormat()...\n");
  
  return ntStatus;
}

void MiniPort::cleanBuffer()
{
  RtlZeroMemory(buffer_, bufferSize_);

  availableBufferToRead_ = bufferSize_; 
}