/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef MiniPort_H       
#define MiniPort_H       

#include <stdunk.h>
#include <portcls.h>
#include <ksdebug.h>

#include "Stream.h"
#include "AudioAdapter.h"
#include "Debug.h"
#include "Topology.h"

//
// Audio buffer size for playback and recording data. Larger buffer 
// means more lattency on recording stream, but more stable recording.
// Too small buffer makes "crackly" recording output.
//

#define BUFFER_SIZE (1024 * 16)

//
// This is 4-byte (little endian dword) Tag for memory alocating.
//

#define NX_POOLTAG '_XN_'

//
// Number of supported streams. This is NOT number of "files", wchich we 
// can play/record at once, becouse sound driver is wrapped by KMIXER.
// Before driver, KMIXER mixes streams into one stream and deliver it
// to our driver if needed.
//

#define MAX_OUTPUT_STREAMS 1
#define MAX_INPUT_STREAMS  1
#define MAX_TOTAL_STREAMS  MAX_OUTPUT_STREAMS + MAX_INPUT_STREAMS                      

//
// Format used internal inside driver
//

#define INTERNAL_CHANNELS    2
#define INTERNAL_RESOLUTION  16
#define INTERNAL_SAMPLE_RATE 44100


class MiniPort :
  public IMiniportWaveCyclic,
  public CUnknown
{
  private:
  
  bool isRecording_;

  bool isPlaying_;
  

  protected:

  AudioAdapter *audioAdapter_;
  
  PPORTWAVECYCLIC portWaveCyclic_;


  static PCFILTER_DESCRIPTOR filterDescriptor_;

  static PCCONNECTION_DESCRIPTOR connections_[];
  
  static PCNODE_DESCRIPTOR nodes_[];
  
  static PCPIN_DESCRIPTOR pins_[];
  

  static KSDATARANGE_AUDIO pinDataRangesStream_[];
  
  static PKSDATARANGE pinDataRangePointersStream_[];
  
  
  static KSDATARANGE pinDataRangesBridge_[];
  
  static PKSDATARANGE pinDataRangePointersBridge_[];
  

  static PCPROPERTY_ITEM propertiesWaveFilter_[];

  static const PCAUTOMATION_TABLE automationWaveFilter_;

  
  ULONG notifyInterval_; 
  
  PSERVICEGROUP serviceGroup_;
  
  KMUTEX sampleRateSync_;
  
  //
  // Common buffer for record and play.
  //
  
  ULONG bufferSize_;
  
  BYTE *buffer_;
  
  int writePos_;

  int readPos_;

  //
  // Variables required to ensure correct order and amount of reading and writing to buffer.
  //
  // writePosLap_ - Idicates how many times pointer writePos_ has wraped to the beginning of the buffer.
  // readPosLap_ - Idicates how many times pointer readPos_ has wraped to the beginning of the buffer.
  // availableBufferToRead_ - Number of bytes that were correctly writen in the buffer and are available to read.
  //

  LONG writePosLap_;

  LONG readPosLap_;

  ULONG availableBufferToRead_; 
  
  //
  // Internal format used by driver.
  //
  
  ULONG channels_;
  ULONG resolution_;
  ULONG sampleRate_;

  
  public:

  DECLARE_STD_UNKNOWN();

  MiniPort(PUNKNOWN unknown);

  ~MiniPort();

  IMP_IMiniportWaveCyclic;

  
  //
  // Property requests handlers.
  //
  
  static NTSTATUS propertyHandler(PPCPROPERTY_REQUEST request);

  NTSTATUS NxPropertyHandler(PPCPROPERTY_REQUEST request);
  
  NTSTATUS propertyHandlerComponentId(PPCPROPERTY_REQUEST request);
  
  NTSTATUS propertyHandlerProposedFormat(PPCPROPERTY_REQUEST request);
    
  NTSTATUS checkFormat(PKSDATAFORMAT dataFormat);
  
  void cleanBuffer();

  
  friend void TimerNotify(PKDPC Dpc, PVOID DeferredContext, PVOID SA1, PVOID SA2);

  friend class Stream;
};


NTSTATUS CreateMiniportWaveCyclic(PUNKNOWN *, REFCLSID, PUNKNOWN, POOL_TYPE);

//
// Pins exposed by Topology.
//                      ------------
// -> WAVEOUT_SRC    -> | TOPOLOGY | -> LINE_OUT_DEST ->
// -> SYNTH_OUT_SRC  -> | NODE     |  
// -> MIC_SOURCE_SRC -> |          | -> WAVE_IN_DEST -> 
//                      ------------

enum
{
  KSPIN_TOPO_WAVEOUT_SOURCE = 0,
  KSPIN_TOPO_SYNTHOUT_SOURCE,
  KSPIN_TOPO_MIC_SOURCE,
  KSPIN_TOPO_LINEOUT_DEST,
  KSPIN_TOPO_WAVEIN_DEST
};

//
// Nodes hidden inside Topology.
//

enum
{
  KSNODE_TOPO_WAVEOUT_VOLUME = 0,
  KSNODE_TOPO_WAVEOUT_MUTE,
  KSNODE_TOPO_SYNTHOUT_VOLUME,
  KSNODE_TOPO_SYNTHOUT_MUTE,
  KSNODE_TOPO_MIC_VOLUME,
  KSNODE_TOPO_LINEOUT_MIX,
  KSNODE_TOPO_LINEOUT_VOLUME,
  KSNODE_TOPO_WAVEIN_MUX,
};

//
// PINs exposed by Wave
//
//                          ---------------
// -> WAVE_RENDER_SINK    -> |WAVE         | -> WAVE_RENDER_SOURCE ->                                                                               
// -> WAVE_CAPTURE_SOURCE -> |    TOPOLOGY | -> WAVE_CAPTURE_SINK ->
//                          ---------------

enum 
{
  KSPIN_WAVE_CAPTURE_SINK = 0,
  KSPIN_WAVE_CAPTURE_SOURCE,
  KSPIN_WAVE_RENDER_SINK, 
  KSPIN_WAVE_RENDER_SOURCE
};

//
// Nodes hidden inside Wave Topology
//

enum 
{
  KSNODE_WAVE_ADC = 0,
  KSNODE_WAVE_DAC
};

//
// Driver info
//

#define NXAUDIO_VERSION  1
#define NXAUDIO_REVISION 0

//
// Guid of custom NX Property set.
//

DEFINE_GUIDSTRUCT("7a16eef7-ebbc-4905-8b8c-25c373a3dbdb", KSPROPSETID_nxaudio);
#define KSPROPSETID_nxaudio DEFINE_GUIDNAMED(KSPROPSETID_nxaudio)

//
// Codes of custom NX properties.
//

enum
{
  KSPROPERTY_NXAUDIO_STATUS = 0
};

#define NXAUDIO_PLAY   1
#define NXAUDIO_RECORD 2

//
// Here are our unique GUIDs generated with guidgen tool, which is
// shipped with visual studio.
//

DEFINE_GUIDSTRUCT("FC010BB1-7D3C-47f4-82DC-107C781C5219", NXAUDIO_PID);
#define NXAUDIO_PID DEFINE_GUIDNAMED(NXAUDIO_PID)

DEFINE_GUIDSTRUCT("7FD8CCE3-2055-4abd-8964-B7070AADE83E", NXAUDIO_NAME);
#define NXAUDIO_NAME DEFINE_GUIDNAMED(NXAUDIO_NAME)

DEFINE_GUIDSTRUCT("C20A81CE-7F53-47a0-9788-C0D2272B9593", NXAUDIO_MANUFACTURER);
#define NXAUDIO_MANUFACTURER DEFINE_GUIDNAMED(NXAUDIO_MANUFACTURER)

#endif
