/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Stream.h"

Stream::Stream(PUNKNOWN outerUnknown) : CUnknown(outerUnknown)
{
  miniPort_         = NULL;
  capture_          = FALSE;
  ksState_          = KSSTATE_STOP;
  blockAlign_       = 0;
  blockAlign_       = 0;

  dmaBuffer_        = NULL;
  dmaPos_           = 0;
  dmaBufferSize_    = 0;
  dmaMovementRate_  = 0;
  dmaTimeStamp_     = 0;
  timeCarryForward_ = 0;
  byteCarryForward_ = 0;

  dpc_              = NULL;
  timer_            = NULL;
}

//
// Destructor sets flags for other stream and frees allocated buffers.
// 

Stream::~Stream(void)
{
  DBG_MSG("Stream::~Stream()...\n");

  if (miniPort_)
  {
    if (capture_)
    {
      miniPort_ -> isRecording_ = false;
    }

    else
    
    {
      miniPort_ -> isPlaying_ = false;
     
      cleanBuffer();
    }
  }

  if (timer_)
  {
    KeCancelTimer(timer_);

    ExFreePoolWithTag(timer_, NX_POOLTAG);
  }

  if (dpc_)
  {
    ExFreePoolWithTag(dpc_, NX_POOLTAG);
  }
}

//
// Initialize new wave stream
// 
// miniPort   - miniport device, wchich will be owner of stream (IN)
// pin        - pin number connected with stream (IN)
// capture    - true, if it's a stream for recordning data (IN)
// dataFormat - format of stream (IN) (WARNING! It's NOT format on PINs) (IN)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Stream::init(MiniPort *miniPort, ULONG pin, 
                          BOOLEAN capture, PKSDATAFORMAT format)
{
  ASSERT(format);

  ASSERT(miniPort);

  UNREFERENCED_PARAMETER(pin);

  DBG_MSG("-> Stream::Init(pin = %u capture = %d)...\n", pin, capture);

  NTSTATUS ntStatus = STATUS_SUCCESS;

  //
  // Fake DMA buffers use one, common buffer allocated in MiniPort.
  //

  miniPort_  = miniPort;

  dmaBuffer_ = miniPort_ -> buffer_;

  dmaBufferSize_ = miniPort_ -> bufferSize_;

  timeCarryForward_ = 0;

  byteCarryForward_ = 0;

  //
  // Set stream format and init state.
  // 

  if (NT_SUCCESS(ntStatus))
  {
    ntStatus = SetFormat(format);
  }

  //
  // Set sample frequency. Note that sampleRateSync_ access should
  // be syncronized.
  //

  if (NT_SUCCESS(ntStatus))
  {
    ksState_ = KSSTATE_STOP;

    capture_ = capture;

    ntStatus = KeWaitForSingleObject(&miniPort_ -> sampleRateSync_,
                                        Executive, KernelMode, FALSE, NULL);

    if (NT_SUCCESS(ntStatus))
    {
      KeReleaseMutex(&miniPort_ -> sampleRateSync_, FALSE);
    }
    else
    {
      DBG_MSG("ERROR. Sync failed.\n");
    }
  }

  if (NT_SUCCESS(ntStatus))
  {
    dpc_ = (PRKDPC) ExAllocatePoolWithTag(NonPagedPool, sizeof(KDPC), NX_POOLTAG);

    if (!dpc_)
    {
      DBG_MSG("ERROR. Cannot allocate memory for dpc.\n");

      ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }
  }

  if (NT_SUCCESS(ntStatus))
  {
    timer_ = (PKTIMER) ExAllocatePoolWithTag(NonPagedPool, sizeof(KTIMER), NX_POOLTAG);

    if (!timer_)
    {
      ExFreePoolWithTag(dpc_, NX_POOLTAG);

      dpc_ = NULL;

      DBG_MSG("ERROR. Cannot allocate memory for timer.\n");

      ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }
  }

  if (NT_SUCCESS(ntStatus))
  {
    KeInitializeDpc(dpc_, TimerNotify, miniPort_);

    KeInitializeTimerEx(timer_, NotificationTimer);
  }
  else
  {
    DBG_MSG("ERROR. Stream::Init() fails (ntStat = %x).\n", ntStatus);
  }

  DBG_MSG("<- Stream::Init()...\n");
  
  return ntStatus;
}

//
// Reads data from stream. Buffer for recording and playback is the same.
// Reading from Buffer is cyclic, i.e. reading from Buffer[MAX + 1]
// means reading from Buffer[1] etc. Both dst and src are kernel - mode
// virtual addresses.
// 
// dst - detination buffer, where copy data from stream (OUT)
// src - pointer to the source buffer, which is located in the DMA
//       channel's buffer (IN)
// len - how many bytes to copy (IN)
//

STDMETHODIMP_(void) Stream::CopyFrom(void *dst, void *src, ULONG len)
{
  if (!dst)
  {
    DBG_MSG("Stream::CopyFrom: ERROR! Invalid input.\n");

    return;
  }

  BYTE *destination = (BYTE *) dst;
    
  //
  // Check how much data in buffer is viable to read and read it.
  //

  if (miniPort_ -> availableBufferToRead_ > 0)
  {
    if (miniPort_ -> availableBufferToRead_ < len)
    {
      len = miniPort_ -> availableBufferToRead_;
    }

    //
    //  Variable aliases only for more clearity.
    //

    int &readPos = miniPort_ -> readPos_;

    int writePos = miniPort_ -> writePos_;

    BYTE *&buffer = miniPort_ -> buffer_;
  
    ULONG &bufferSize = miniPort_ -> bufferSize_;

    ULONG real = len;

    //
    // Read data from buffer.
    //

    if ((readPos + real) < bufferSize)
    {
      if (writePos > readPos)
      {
        if (readPos + len > (ULONG) writePos)
        {
          real = writePos - readPos;
        }
      }

      RtlZeroMemory(destination, sizeof(destination));

      RtlCopyMemory(destination, buffer + readPos, real);

      miniPort_ -> availableBufferToRead_ -= real; 

      if (real < (ULONG) len)
      {
        miniPort_ -> availableBufferToRead_ -= (len - real);
      }
    }
    else
    {
      ULONG bytesToEnd = bufferSize - readPos;
    
      ULONG overflow = len - bytesToEnd;
    
      if (overflow >= (ULONG) readPos)
      {
        overflow -= readPos;
      }

      if (overflow >= (ULONG) writePos)
      {
        overflow -= writePos;
      }

      real = bytesToEnd + overflow;

      RtlZeroMemory(destination, sizeof(destination));

      RtlCopyMemory(destination, buffer + readPos, bytesToEnd);

      RtlCopyMemory(destination + bytesToEnd, buffer, overflow);

      miniPort_ -> availableBufferToRead_ -= real;

      if (real < (ULONG) len)
      {
        miniPort_ -> availableBufferToRead_ -= (len - real);
      }

      changeLapRead();
    }

    readPos = (readPos + real) % bufferSize;
  }

  UNREFERENCED_PARAMETER(src);
}

//
// Write data to stream. Buffer for recording and playback is the same.
// Writing to Buffer is cyclic, i.e. writing to Buffer[MAX + 1]
// means writing to Buffer[1] etc. Both dst and src are kernel - mode
// virtual addresses.
// 
// dst - not used
// src - source buffer (IN)
// len - how many bytes to copy (IN)
//

STDMETHODIMP_(void) Stream::CopyTo(void *dst, void *src, ULONG len)
{
  if (!src)
  {
    DBG_MSG("Stream::CopyFrom: ERROR! Invalid input.\n");

    return;
  }

  BYTE *source = (BYTE *) src;

  //
  // Variale aliases only for more clearity.
  //

  LONG &writePosLap = miniPort_ -> writePosLap_;

  LONG &readPosLap = miniPort_ -> readPosLap_;

  int &writePos = miniPort_ -> writePos_;

  int &readPos = miniPort_ -> readPos_;

  BYTE *&buffer = miniPort_ -> buffer_;

  ULONG &bufferSize = miniPort_ -> bufferSize_;

  // 
  // Check if writing data will not overwrite data not red.
  // If so write until readPos was reached.
  //

  if (writePosLap == readPosLap)
  {
    if ((writePos + len) % bufferSize > (ULONG) readPos)
    { 
      if (len > (bufferSize - writePos) + readPos)
      {
        len = (bufferSize - writePos) + readPos;
      }
    }
  }

  else
  {
    if (len >= (ULONG) readPos - writePos)
    {
      len = readPos - writePos;
    }
  }

  miniPort_ -> availableBufferToRead_ += len;

  int real = len;
  
  //
  // Write data to buffer.
  //
  
  if ((writePos + len) < bufferSize)
  {
    RtlCopyMemory(buffer + writePos, source, len);
  }
  else
  {
    ULONG bytesToEnd = bufferSize - writePos;
    
    ULONG overflow = len - bytesToEnd;

    if (overflow >=(ULONG) writePos)
    {
      overflow = writePos - 1;

      real = overflow;
    }
    
    RtlCopyMemory(buffer + writePos, source, bytesToEnd);

    RtlCopyMemory(buffer, source + bytesToEnd, overflow);

    changeLapWrite();
  }

  writePos = (writePos + real) % bufferSize;

  UNREFERENCED_PARAMETER(dst);
}

//
// Clean up stream buffer.
//

void Stream::cleanBuffer()
{
  miniPort_ -> cleanBuffer();
}

//
// Callback function for sample rate synchronization.
//

void TimerNotify(PKDPC dpc, PVOID context, PVOID sa1, PVOID sa2)
{
  UNREFERENCED_PARAMETER(dpc);
  UNREFERENCED_PARAMETER(sa1);
  UNREFERENCED_PARAMETER(sa2);

  MiniPort *miniport = (MiniPort *) context;

  if (miniport && miniport -> portWaveCyclic_)
  {
    miniport -> portWaveCyclic_ -> Notify(miniport -> serviceGroup_);
  }
}

//
// Set new stream's format
//
// format - new format what we want to set (IN)
// 
// RETURNS: NTSTATUS code.
//

STDMETHODIMP_(NTSTATUS) Stream::SetFormat(PKSDATAFORMAT format)
{
  DBG_MSG("-> Stream::SetFormat()...\n");

  NTSTATUS ntStatus = STATUS_INVALID_DEVICE_REQUEST;

  PWAVEFORMATEX pWfx = NULL;

  
  FAIL(format == NULL);
  
  FAIL(ksState_ == KSSTATE_RUN);
  
  //
  // Xp may try set 'bad' format even if it is not listed
  // in driver initialization. We validate format once again here.
  //
  
  if (miniPort_)
  {
    NTFAIL((miniPort_ -> checkFormat(format)));
  }  
  
  //
  // Get extended format.
  // 
  
  if (format -> Specifier == KSDATAFORMAT_SPECIFIER_DSOUND)
  {
    pWfx = &PKSDSOUND_BUFFERDESC(format + 1) -> WaveFormatEx;
  }
  else
  {
    pWfx = PWAVEFORMATEX(format + 1);
  }

  //
  // Set stream format.
  //

  NTFAIL(KeWaitForSingleObject(&miniPort_ -> sampleRateSync_,
                                   Executive, KernelMode, FALSE, NULL));
  
  blockAlign_ = pWfx -> nBlockAlign;

  dmaMovementRate_ = pWfx -> nAvgBytesPerSec;

  KeReleaseMutex(&miniPort_ -> sampleRateSync_, FALSE);

  DBG_MSG("New Format: %d\n", pWfx -> nSamplesPerSec);  

  ntStatus = STATUS_SUCCESS;
  
fail:  
  
  if (ntStatus)
  {
    DBG_MSG("ERROR. Stream::SetFormat() fails (ntStat = %x).\n", ntStatus);
  }

  DBG_MSG("<- Stream::SetFormat()...\n");
  
  return ntStatus;
}


//
// Set interrupt time interval and compute time interval in byte 
// (frame size).
//
// interval  - time interval in ms (IN)
// frameSize - how much bytes are transfer in given interval (OUT)
//
// RETURNS: NTSTATUS code.
//

STDMETHODIMP_(ULONG) Stream::SetNotificationFreq(ULONG interval, PULONG frameSize)
{
  miniPort_-> notifyInterval_ = interval;

  *frameSize = blockAlign_ * miniPort_ -> sampleRate_ * interval / 1000;

  return miniPort_ -> notifyInterval_;
}


//
// Change stream state (stop, run, pause etc).
//
// state - new state to set (IN)
//

STDMETHODIMP Stream::SetState(KSSTATE state)
{
  DBG_MSG("Stream::SetState()...\n");

  NTSTATUS ntStatus = STATUS_SUCCESS;

  
  if (state == KSSTATE_ACQUIRE)
  {
    state = KSSTATE_STOP;
  }

  //
  // If state changed.
  //
  
  if (state != ksState_)
  {
    switch(state)
    {
      //
      // Pause.
      //
      
      case KSSTATE_PAUSE:
      {
        DBG_MSG("KSSTATE_PAUSE\n");

        cleanBuffer();
 
        break;
      }

      //
      // Run.
      //
      
      case KSSTATE_RUN:
      {
        DBG_MSG("KSSTATE_RUN\n");

        LARGE_INTEGER delay = {0};

        //
        // Set the timer for DPC.
        //
        
        dmaTimeStamp_ = KeQueryInterruptTime();
        
        timeCarryForward_ = 0;
        
        delay.HighPart = 0;

        delay.LowPart = miniPort_ -> notifyInterval_;

        KeSetTimerEx(timer_, delay, miniPort_ -> notifyInterval_, dpc_);

        break;
      }

      //
      // Stop.
      //
      
      case KSSTATE_STOP:
      {

        DBG_MSG("KSSTATE_STOP\n");

        dmaPos_ = 0;
        
        timeCarryForward_ = 0;
        
        byteCarryForward_ = 0;

        KeCancelTimer(timer_);

        break;
      }
    }
    
    ksState_ = state;
  }

  return ntStatus;
}


//
// Fill buffer with null audio signal. In our case it is zero fill (PCM).
//
// buffer - buffer to fill (OUT)
// size   - how many bytes to fill (IN)
//

STDMETHODIMP_(void) Stream::Silence(PVOID buffer, ULONG size)
{
  miniPort_ -> availableBufferToRead_ += size;

  RtlZeroMemory(buffer, size);
}

void Stream::changeLapRead ()
{
  if (miniPort_ -> readPosLap_ >= LONG_MAX )
  {
    miniPort_ -> readPosLap_ = 0;
  }
  else
  {
    miniPort_ -> readPosLap_ ++;
  }
}

void Stream::changeLapWrite ()
{
  if (miniPort_ -> writePosLap_ >= LONG_MAX )
  {
    miniPort_ -> writePosLap_ = 0;
  }
  else
  {
    miniPort_ -> writePosLap_ ++;
  }
}


//
// Below functions implement IDMACHANNEL interface. 
//

STDMETHODIMP Stream::GetPosition(PULONG pos)
{
  if (ksState_ == KSSTATE_RUN)
  {
    //
    // Get the current time
    //

    ULONGLONG currentTime = KeQueryInterruptTime();

    // Calculate the time elapsed since the last call to GetPosition() or since the
    // DMA engine started.  Note that the division by 10000 to convert to milliseconds
    // may cause us to lose some of the time, so we will carry the remainder forward 
    // to the next GetPosition() call.
    //

    ULONG timeElapsedInMs = ((ULONG) (currentTime - dmaTimeStamp_ + timeCarryForward_)) / 10000;

    //
    // Carry forward the remainder of this division so we don't fall behind with our position.
    //
    
    timeCarryForward_ = (currentTime - dmaTimeStamp_ + timeCarryForward_) % 10000;

    //
    // Calculate how many bytes in the DMA buffer would have been processed in the elapsed
    // time.  Note that the division by 1000 to convert to milliseconds may cause us to
    // lose some bytes, so we will carry the remainder forward to the next GetPosition() call.
    //
    
    ULONG byteDisplacement = ((dmaMovementRate_ * timeElapsedInMs) + byteCarryForward_) / 1000;

    //
    // Carry forward the remainder of this division so we don't fall behind with our position.
    //
    
    byteCarryForward_ = ((dmaMovementRate_ * timeElapsedInMs) + byteCarryForward_) % 1000;

    //
    // Increment the DMA position by the number of bytes displaced since the last
    // call to GetPosition() and ensure we properly wrap at buffer length.
    //
    
    dmaPos_ = (dmaPos_ + byteDisplacement) % dmaBufferSize_;

    //
    // Return the new DMA position
    //
    
    *pos = dmaPos_;

    //
    // Update the DMA time stamp for the next call to GetPosition()
    //
    
    dmaTimeStamp_ = currentTime;
  }
  else
  {
    //
    // DMA is inactive so just return the current DMA position.
    //
    
    *pos = dmaPos_;
  }

  return STATUS_SUCCESS;
}


STDMETHODIMP Stream::NormalizePhysicalPosition(PLONGLONG pos)
{
  *pos = (_100NS_UNITS_PER_SECOND / blockAlign_ * *pos) / miniPort_ -> sampleRate_;

  return STATUS_SUCCESS;
}


//
// For compatibility only. All below functions implement fake DMA 
// buffer. In the real, all our streams (for record and playback) use
// one, common buffer alocated in MiniPort object.
//

STDMETHODIMP_(ULONG) Stream::BufferSize(void)
{
  return dmaBufferSize_;
}

STDMETHODIMP_(ULONG) Stream::AllocatedBufferSize()
{
  return dmaBufferSize_;
}


STDMETHODIMP_(void) Stream::FreeBuffer()
{
}


STDMETHODIMP_(ULONG) Stream::MaximumBufferSize()
{
  return miniPort_ -> bufferSize_;
}


STDMETHODIMP_(PHYSICAL_ADDRESS) Stream::PhysicalAddress()
{
  PHYSICAL_ADDRESS address = {0};

  address.QuadPart = (LONGLONG) dmaBuffer_;

  return address;
}


STDMETHODIMP_(void) Stream::SetBufferSize(ULONG size)
{
  if (size <= dmaBufferSize_)
  {
    dmaBufferSize_ = size;
  }
}


STDMETHODIMP_(PVOID) Stream::SystemAddress()
{
  return dmaBuffer_;
}


STDMETHODIMP_(ULONG) Stream::TransferCount()
{
  return dmaBufferSize_;
}


STDMETHODIMP_(PADAPTER_OBJECT) Stream::GetAdapterObject()
{
  return NULL;
}


STDMETHODIMP_(NTSTATUS) Stream::AllocateBuffer(ULONG size, PPHYSICAL_ADDRESS constraint)
{
  UNREFERENCED_PARAMETER(constraint);

  dmaBuffer_ = miniPort_ -> buffer_;

  dmaBufferSize_ = size;

  return STATUS_SUCCESS;
}

//
// Query one of interface implemented by class (COM model) and 
// increese its reference counter if success.
//
// id     - interface id (IN)
// object - pointer to quered interface (OUT)
//
// RETURNS: NTSTATUS code.
//

STDMETHODIMP_(NTSTATUS) Stream::NonDelegatingQueryInterface(REFIID id, PVOID *obj)
{
  DBG_MSG("UknownStream::NonDelegatingQueryInterface()...\n");

  ASSERT(obj);

  if (id == IID_IUnknown)
  {
    *obj = PUNKNOWN(PMINIPORTWAVECYCLICSTREAM(this));
  }
  else if (id == IID_IMiniportWaveCyclicStream)
  {
    *obj = (PMINIPORTWAVECYCLICSTREAM) this;
  }
  else if (id == IID_IDmaChannel)
  {
    *obj = (PDMACHANNEL) this;
  }
  else
  {
    *obj = NULL;
  }

  if (*obj)
  {
    PUNKNOWN(*obj) -> AddRef();

    return STATUS_SUCCESS;
  }


  return CUnknown::NonDelegatingQueryInterface(id, obj);
}