/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2011, 2019 NoMachine, http://www.nomachine.com/.         */
/*                                                                        */
/* NXAUDIO, NX protocol compression and extensions to this software       */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#include "Topology.h"

#pragma code_seg("PAGE")

Topology::Topology(PUNKNOWN outer) : CUnknown(outer)
{
  PAGED_CODE();

  audioAdapter_ = NULL;
}

//
// Create Topology object for driver initialization code (-> 'Init.cpp').
//

NTSTATUS CreateMiniPortTopology(PUNKNOWN *unknown, REFCLSID,
                                    PUNKNOWN outer, POOL_TYPE poolType)
{
  PAGED_CODE();

  DBG_MSG("CreateMiniportTopology()...\n");

  ASSERT(unknown);

  STD_CREATE_BODY(Topology, unknown, outer, poolType);
}


//
// Initialize new Topology object.
//
// unknown   - pointer to CUnknown object (IN)
// resources - resources list assigned by PnP manager (IN)
// port      - mother PortTopology object (IN)
//

STDMETHODIMP Topology::Init(PUNKNOWN unknown, 
                                PRESOURCELIST resources, PPORTTOPOLOGY port)
{
  DBG_MSG("Topology::Init()...\n");
  
  PAGED_CODE();
  
  UNREFERENCED_PARAMETER(resources);
  
  UNREFERENCED_PARAMETER(port);

  NTSTATUS ntStatus = STATUS_UNSUCCESSFUL;
  
 
  NTFAIL(unknown -> QueryInterface(IID_AudioAdapter, (PVOID *) &audioAdapter_));

  audioAdapter_ -> mixerReset();

  audioAdapter_ -> setMux(KSPIN_TOPO_MIC_SOURCE);

  ntStatus = STATUS_SUCCESS;
  
fail:

  //
  // Clean up on error.
  //
  
  if (ntStatus)
  {
    DBG_MSG("ERROR. Topology::Init() fails (%x).\n", ntStatus);

    //
    // clean up AdapterCommon
    //
    
    if (audioAdapter_)
    {
      audioAdapter_ -> Release();

      audioAdapter_ = NULL;
    }    
  }
  
  return ntStatus;
}


//
// Handler for Topology Filter Property Requests.
//
// For compatibility only.
//

NTSTATUS Topology::propertyHandlerFilter(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();

  UNREFERENCED_PARAMETER(request);
  
  DBG_MSG("Topology::propertyHandlerFilter");

  return STATUS_NOT_IMPLEMENTED;
}

//
// General handler for basic support request.
// 
// request     - structure with request's description (IN/OUT).
// accessFlags - property access flag to set (e.g SET + GET) (IN).
// propType    - property type (e.g. VT_BOOL) (IN).
//
// RETURNS: 0 if OK.
//

NTSTATUS Topology::propertyHandlerBasicSupport(PPCPROPERTY_REQUEST request,
                                                   ULONG accessFlags, DWORD propType)
{
  PAGED_CODE();
  
  DBG_MSG("Topology::propertyHandlerBasicSupport()...\n");
  
  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;
  
  //
  // Check args.
  //
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);
  
  if (request -> ValueSize < sizeof(ULONG))
  {
    request -> ValueSize = 0;
    
    NTFAIL(STATUS_BUFFER_TOO_SMALL);
  }

  //
  // Do we can put full PROPERTY_DESCRIPTION? If yes, we should do it.
  //
  
  if (request -> ValueSize >= sizeof(KSPROPERTY_DESCRIPTION))
  {
    PKSPROPERTY_DESCRIPTION desc = PKSPROPERTY_DESCRIPTION(request -> Value);

    desc -> AccessFlags     = accessFlags;
    desc -> DescriptionSize = sizeof(KSPROPERTY_DESCRIPTION);
    
    if (VT_ILLEGAL != propType)
    {
      desc -> PropTypeSet.Set = KSPROPTYPESETID_General;
      desc -> PropTypeSet.Id  = propType;
    }
    else
    {
      desc -> PropTypeSet.Set = GUID_NULL;
      desc -> PropTypeSet.Id  = 0;
    }

    desc -> PropTypeSet.Flags = 0;
    desc -> MembersListCount  = 0;
    desc -> Reserved          = 0;

    request -> ValueSize = sizeof(KSPROPERTY_DESCRIPTION);
  } 
  
  //
  // If not, we put only access flags, to inform system what he can do
  // with given properties (get, set etc.).
  //
  
  else  
  {
    *(PULONG(request -> Value)) = accessFlags;

    request -> ValueSize = sizeof(ULONG);
  }

  //
  // Error handler.
  //
  
  ntStatus = STATUS_SUCCESS;
  
  fail:
  
  if (ntStatus)
  {
    DBG_MSG("ERROR. Topology::propertyHandlerBasicSupport() fails."
                " NtStatus = 0x%x\n", ntStatus);
  }
  
  return ntStatus;
}

//
// Handler for all topology properties requests.
//
// request - structure describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Topology::propertyHandler(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();
  
  NTSTATUS ntStatus = STATUS_INVALID_DEVICE_REQUEST;
  
  DBG_MSG("Topology::propertyHandler [id = %d]", request -> PropertyItem -> Id);

  //
  // Get target Topology object.
  //
  
  Topology *target = (Topology *) (request -> MajorTarget);
    
  //
  // Redirect requests to functions in target object.
  //

  switch (request -> PropertyItem -> Id) 
  {
    //
    // Volume properties.
    //

    case KSPROPERTY_AUDIO_VOLUMELEVEL:
    {
      ntStatus = target -> propertyHandlerVolume(request);
     
      break;
    }

    //
    // Mute properties.
    //

    case KSPROPERTY_AUDIO_MUTE:
    {
      ntStatus = target -> propertyHandlerMute(request);
      
      break;
    }

    //
    // Source mux properties.
    //
    
    case KSPROPERTY_AUDIO_MUX_SOURCE:
    {
      ntStatus = target -> propertyHandlerMuxSource(request);
     
      break;
    }

    //
    // System asks us does node is implemented in hardware or software.
    //
    
    case KSPROPERTY_AUDIO_CPU_RESOURCES:
    {
      ntStatus = target -> propertyHandlerCpuResource(request);
      
      break;
    }

    default:
    {
      DBG_MSG("Topology::propertyHandler: Invalid Device Request [%d]", 
                  request -> PropertyItem -> Id);
    }
  }
  
  return ntStatus;
}


//
// Handler for get/set/basic mute property.
//
// request - struct describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Topology::propertyHandlerMute(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();

  DBG_MSG("Topology::propertyHandlerMute()...\n");

  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;

  //
  // Check arguments.
  //
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);

  //
  // Property basic support request.
  //

  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    NTFAIL(propertyHandlerBasicSupport(request, KSPROPERTY_TYPE_ALL, VT_BOOL));
  }
  else
  {
    BOOL *pValue = (BOOL *) request -> Value;
    
    if (request -> ValueSize < sizeof(BOOL))
    {
      request -> ValueSize = 0;
      
      NTFAIL(STATUS_BUFFER_TOO_SMALL);
    }

    //
    // Property get request.
    //
    
    if (request -> Verb & KSPROPERTY_TYPE_GET)
    {
      *pValue = audioAdapter_ -> getMute(request -> Node);
    
      request -> ValueSize = sizeof(BOOL);
    }

    //
    // Property set request.
    //
    
    else if (request -> Verb & KSPROPERTY_TYPE_SET)
    {
      audioAdapter_ -> setMute(request -> Node, *pValue);
    }
  }

  ntStatus = STATUS_SUCCESS;
  
fail:  
  
  if (ntStatus)
  {
    DBG_MSG("ERROR. propertyHandlerMute fails (%x).\n", ntStatus);
  }

  return ntStatus;
}


//
// Handler for get/set/basic source mux property.
//
// request - struct describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Topology::propertyHandlerMuxSource(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();

  NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;

  //
  // Check arguments.
  //

  FAIL(request == NULL);
  FAIL(request -> Value == NULL);

  if (request -> ValueSize < sizeof(ULONG))
  {
    request -> ValueSize = 0;
   
    NTFAIL(STATUS_BUFFER_TOO_SMALL);
  }

  ULONG *pValue = (ULONG *) request -> Value;

  //
  // Basic support request.
  //
  
  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    NTFAIL(propertyHandlerBasicSupport(request, KSPROPERTY_TYPE_ALL, VT_I4));
  }

  //
  // Get property request. 
  //
  
  else if (request -> Verb & KSPROPERTY_TYPE_GET)
  {
    *pValue = audioAdapter_ -> getMux();
    
    request -> ValueSize = sizeof(ULONG);
  }

  //
  // Set property request.
  //
  
  else if (request -> Verb & KSPROPERTY_TYPE_SET)
  {
    audioAdapter_ -> setMux(*pValue);
  }

  ntStatus = STATUS_SUCCESS;
  
fail:

  if (ntStatus)
  {
    DBG_MSG("ERROR. propertyHandlerMuxSource() fails (%x).\n", ntStatus);
  }
  
  return ntStatus;
}


//
// Handler for get/set/basic volume property.
//
// request - struct describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Topology::propertyHandlerVolume(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();

  DBG_MSG("Topology::propertyHandlerVolume()...\n");
            
  NTSTATUS ntStatus = STATUS_INVALID_DEVICE_REQUEST;
  
  //
  // Check arguments.
  //
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);
  
  if (request -> ValueSize < sizeof(ULONG))
  {
    NTFAIL(STATUS_BUFFER_TOO_SMALL);
  }

  //
  // Basic request. We can't use general handler here, becouse we want
  // to send possible volume range. It's specified for volume property
  // only.
  //
  //
  
  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    ULONG cbFullProperty = sizeof(KSPROPERTY_DESCRIPTION) 
                         + sizeof(KSPROPERTY_MEMBERSHEADER)
                         + sizeof(KSPROPERTY_STEPPING_LONG);

    if (request -> ValueSize >= (sizeof(KSPROPERTY_DESCRIPTION)))
    {
      PKSPROPERTY_DESCRIPTION desc = PKSPROPERTY_DESCRIPTION(request -> Value);

      desc -> AccessFlags       = KSPROPERTY_TYPE_ALL;
      desc -> DescriptionSize   = cbFullProperty;
      desc -> PropTypeSet.Set   = KSPROPTYPESETID_General;
      desc -> PropTypeSet.Id    = VT_I4;
      desc -> PropTypeSet.Flags = 0;
      desc -> MembersListCount  = 1;
      desc -> Reserved          = 0;

      //
      // If buffer can also hold a volume range we should pass it.
      //
        
      if(request -> ValueSize >= cbFullProperty)
      {
        //
        // Members header.
        //
            
        PKSPROPERTY_MEMBERSHEADER members = PKSPROPERTY_MEMBERSHEADER(desc + 1);

        members -> MembersFlags = KSPROPERTY_MEMBER_STEPPEDRANGES;
        members -> MembersSize  = sizeof(KSPROPERTY_STEPPING_LONG);
        members -> MembersCount = 1;
        members -> Flags        = KSPROPERTY_MEMBER_FLAG_BASICSUPPORT_MULTICHANNEL;

        //
        // Possible volume range and minimal step.
        //
            
        PKSPROPERTY_STEPPING_LONG range = PKSPROPERTY_STEPPING_LONG(members + 1);

        range -> Bounds.SignedMaximum = 0x00000000;    //   0 dB
        range -> Bounds.SignedMinimum = -96 * 0x10000; // -96 dB
        range -> SteppingDelta        = 0x08000;       //  .5 dB
        range -> Reserved             = 0;             //

        //
        // set the return value size.
        //
            
        request -> ValueSize = cbFullProperty;
      } 
      else
      {
        request -> ValueSize = sizeof(KSPROPERTY_DESCRIPTION);
      }
    } 
    else
    {
      //
      // if return buffer can hold only a ULONG, return the access flags.
      // This is the same as general handler code.
      //
      
      PULONG accessFlags = PULONG(request -> Value);

      request -> ValueSize = sizeof(ULONG);

      *accessFlags = KSPROPERTY_TYPE_ALL;
    }
  }
  
  //
  // Non basic requests.
  //
  
  else
  {
    ULONG *pVolume = (ULONG *) request -> Value;
    
    //
    // Get property request.
    //
    
    if (request -> Verb & KSPROPERTY_TYPE_GET)
    {
      *pVolume = audioAdapter_ -> getVolume(request -> Node);

      request -> ValueSize = sizeof(ULONG);
    }
    
    //
    // Set property request.
    //
    
    else if (request -> Verb & KSPROPERTY_TYPE_SET)
    {
      audioAdapter_ -> setVolume(request -> Node, *pVolume);
    }
    
    //
    // Unknown request type.
    //
    
    else
    {
      NTFAIL(STATUS_INVALID_PARAMETER);
    }
  }

  ntStatus = STATUS_SUCCESS;

fail:

  if (ntStatus)
  {
    DBG_MSG("ERROR. propertyHandleVolume() fails (%x).\n", ntStatus);
  }
  
  return ntStatus;
}

//
// Handler for get/set/basic cpu_resource property.
//
// request - struct describing request (IN/OUT)
//
// RETURNS: NTSTATUS code.
//

NTSTATUS Topology::propertyHandlerCpuResource(PPCPROPERTY_REQUEST request)
{
  PAGED_CODE();

  DBG_MSG("Topology::propertyHandlerCpuResource()...\n");
            
  NTSTATUS ntStatus = STATUS_INVALID_DEVICE_REQUEST;
  
  //
  // Check arguments.
  //
  
  FAIL(request == NULL);
  FAIL(request -> Value == NULL);
  
  if (request -> ValueSize < sizeof(ULONG))
  {
    NTFAIL(STATUS_BUFFER_TOO_SMALL);
  }

  //
  // Basic request. We allow only GET and BasicSupport requests for 
  // this property.
  //

  if (request -> Verb & KSPROPERTY_TYPE_BASICSUPPORT)
  {
    NTFAIL(propertyHandlerBasicSupport(request, KSPROPERTY_TYPE_GET
                                           | KSPROPERTY_TYPE_BASICSUPPORT,
                                               VT_ILLEGAL));
  }
  
  //
  // Get request.
  //
  
  else if (request -> Verb & KSPROPERTY_TYPE_GET)
  {
    *(PULONG(request -> Value)) = KSAUDIO_CPU_RESOURCES_NOT_HOST_CPU;

    request -> ValueSize = sizeof(ULONG);
  }
  else
  {
    NTFAIL(STATUS_INVALID_PARAMETER);
  }

  ntStatus = STATUS_SUCCESS;

fail:

  if (ntStatus)
  {
    DBG_MSG("ERROR. propertyHandleVolume() fails (%x).\n", ntStatus);
  }
  
  return ntStatus;
}


//
// Descructor decrease reference counter for used interface (COM model).
//

Topology::~Topology(void)
{
  PAGED_CODE();

  DBG_MSG("Topology::~Topology()...\n");
  
  if (audioAdapter_)
  {
    audioAdapter_ -> Release();
  }
}

//
// Retrieve format range supported on given pin.
//
// Unused. Only for compatibility.
//

NTSTATUS Topology::DataRangeIntersection(ULONG pinId, 
                                             PKSDATARANGE clientDataRange,
                                                 PKSDATARANGE myDataRange,
                                                     ULONG outputBufferLength,
                                                         PVOID resultantFormat,
                                                             PULONG resultantFormatLength)
{
  PAGED_CODE();

  DBG_MSG("Topology::DataRangeIntersection()...\n");
  
  UNREFERENCED_PARAMETER(pinId);
  UNREFERENCED_PARAMETER(clientDataRange);
  UNREFERENCED_PARAMETER(myDataRange);
  UNREFERENCED_PARAMETER(outputBufferLength);
  UNREFERENCED_PARAMETER(resultantFormat);
  UNREFERENCED_PARAMETER(resultantFormatLength);

  return STATUS_NOT_IMPLEMENTED;
}


//
// Retrieve filter descriptor table (-> see TopologyTable.cpp).
//
// filterDescriptor - pointer to table (OUT)
// 
// RETURNS: NTSTATUS code.
//

STDMETHODIMP Topology::GetDescription(PPCFILTER_DESCRIPTOR *filterDescriptor)
{
  PAGED_CODE();

  DBG_MSG("Topology::GetDescription()...\n");
  
  *filterDescriptor = &filterDescriptor_;

  return STATUS_SUCCESS;
}


// Query one of interfaces implemented by class (COM model) and 
// increese its reference counter if success.
//
// id     - interface id (IN)
// object - pointer to quered interface (OUT)
//
// RETURNS: NTSTATUS code.

STDMETHODIMP Topology::NonDelegatingQueryInterface(REFIID id, PVOID *object)
{
  PAGED_CODE();

  ASSERT(object);

  DBG_MSG("Topology::NonDelegatingQueryInterface()...\n");

  if (id == IID_IUnknown)
  {
    *object = (PUNKNOWN) this;
  }
  else if (id == IID_IMiniport)
  {
    *object = (PMINIPORT) this;
  }
  else if (id == IID_IMiniportTopology)
  {
    *object = (PMINIPORTTOPOLOGY) this;
  }
  else
  {
    *object = NULL;
  }

  //
  // Increse reference counter for given interface.
  //

  if (*object)
  {
    PUNKNOWN(*object) -> AddRef();

    return STATUS_SUCCESS;
  }

  return CUnknown::NonDelegatingQueryInterface(id, object);
}

#pragma code_seg()